# ncgmp09_ValidateDatabase.py  
#   Python script to inventory tables, feature datasets
#   and feature classes in a geodatabase and to check
#   for conformance with NCGMP09 geodatabase schema.
#   For more information, see tail of this file. 
#   Assumes ArcGIS 9.3 or higher.
#   Ralph Haugerud, USGS
#
#   Takes two arguments: <geodatabaseName> <outputWorkspace>
#     and writes a file named <geodatabaseName>-NCGMP09conformance.txt.
#   At present only works on a geodatabase in the local directory.
#   Requires that ncgmp09_definition.py be present in the local directory 
#     or in the appropriate Python library directory.
#   Incomplete error trapping and no input checking. 
#
# NOTE: THIS CODE HAS NOT BEEN THOROUGHLY TESTED
#   If you find problems, or think you see errors, please let me know.
#   Zip up the database, the conformance report (if one is written), 
#   a brief discussion of what's wrong, and email zip file to me at
#   	rhaugerud@usgs.gov
#   Please include "NCGMP09" in the subject line.
#   Thanks!

import arcgisscripting, sys, time, os.path
from ncgmp09_definition import tableDict 

versionString = 'ngcmp09_ValidateDatabase.py, version of 6 October 2009'

debug = False

# fields we don't want listed or described when inventorying dataset:
standardFields = ('OBJECTID','SHAPE','Shape','SHAPE_Length','SHAPE_Area','ZOrder',
                  'AnnotationClassID','Status','TextString','FontName','FontSize','Bold',
                  'Italic','Underline','VerticalAlignment','HorizontalAlignment',
                  'XOffset','YOffset','Angle','FontLeading','WordSpacing','CharacterWidth',
			'CharacterSpacing','FlipAngle','Override','Shape_Length','Shape_Area') 

# fields whose values must be defined in Glossary
#   Note that PropertyValue terms that contain numbers need not be defined
gFieldDefList = ('Type','LocationMethod','PartType','ProportionTerm','TimeScale','Qualifier',
		     'Property','PropertyValue','ExistenceConfidence','IdentityConfidence',
		     'ScientificConfidence','ParagraphStyle','AgeUnits')

tables = []
fdsfc = []
all_IDs = []		# list of should-be unique identifiers
allMapUnitRefs = []	# list of MapUnit references (from various Poly feature data sets)
allGlossaryRefs = []	# list of all references to Glossary
allDataSourcesRefs = [] # list of all references to DataSources
allBadNulls = ["""  Pseudonulls (value = <space>) commonly result from loading empty data into
  string fields in which nulls are not allowed. Trailing spaces are commonly
  produced by hand-correction of pseudonulls. The following fields contain
  pseudonulls or trailing spaces"""]

gdbDescription = []
schemaErrors = []
schemaExtensions = []

duplicateIDs = []
unreferencedIds = ['  OwnerIDs and ValueLinkIDs in ExtendedAttributes that are absent elsewhere in the database']
extendedAttribIDs = []
missingSourceIDs = ['  Missing DataSources entries. Only one reference to each missing source is cited']
unusedDataSources = ['  Entries in DataSources that are not otherwise referenced in database']
dataSourcesIDs = []
missingDmuMapUnits = ['  MapUnits missing from DMU. Only one reference to each missing unit is cited']
missingStandardLithMapUnits = ['  MapUnits missing from StandardLithology. Only one reference to each missing unit is cited']
unreferencedDmuMapUnits = ['  MapUnits in DMU that are not present on map']
unreferencedStandardLithMapUnits = ['  MapUnits in StandardLithology that are not present on map']
dmuMapUnits = []
standardLithMapUnits = []
missingGlossaryTerms = ['  Missing terms in Glossary. Only one reference to each missing term is cited']
unusedGlossaryTerms = ['  Terms in Glossary that are not otherwise used in geodatabase']
glossaryTerms = []
unusedGeologicEvents = ['  Events in GeologicEvents that are not cited in ExtendedAttributes']
hKeyErrors = ['  HierarchyKey errors, DescriptionOfMapUnits']

def addMsgAndPrint(gp,msg, severity=0): 
	# prints msg to screen and adds msg to the geoprocessor (in case this is run as a tool) 
	# print msg 
	try: 
	  for string in msg.split('\n'): 
		# Add appropriate geoprocessing message 
		if severity == 0: 
			gp.AddMessage(string) 
		elif severity == 1: 
			gp.AddWarning(string) 
		elif severity == 2: 
			gp.AddError(string) 
	except: 
		pass 

def writeContentErrors(outfl,errors,noErrorString):
	if len(errors) == 1:
		outfl.write('  '+noErrorString+'\n\n')
	else: 
		for aline in errors:
			outfl.write(aline+'\n')
		outfl.write('\n')

def listDataSet(dataSet):
	result = gp.GetCount_management(dataSet)
	nrows = int(result.GetOutput(0))
	gdbDescription.append('    '+dataSet+', '+str(nrows)+' records')
	fields = gp.ListFields(dataSet)
	for field in fields:
	   if not (field.Name in standardFields):
		gdbDescription.append('      '+field.Name+' '+field.Type+':'+str(field.Length)+'  '+str(field.Required))

def checkMapFeatureClasses(gp,fds,prefix,fcs):
	addMsgAndPrint(gp,'  Checking for required feature classes...')
	requiredFeatureClasses = []
	for fc in ['ContactsAndFaults','MapUnitPolys']:
		requiredFeatureClasses.append(prefix+fc)
	for fc in requiredFeatureClasses:
		if not (fc in fcs): 
			schemaErrors.append('Feature data set '+fds+', feature class '+fc+' is missing')

def checkTableFields(dBTable,defTable):
	dBtable = str(dBTable)
	# build dictionary of required fields 
	requiredFields = {}
	requiredFieldDefs = tableDict[defTable]
	for fieldDef in requiredFieldDefs:
		requiredFields[fieldDef[0]] = fieldDef
	# build dictionary of existing fields 
	existingFields = {}
	fields = gp.ListFields(dBtable)
	for field in fields:
		existingFields[field.Name] = field
	# now check to see what is excess / missing
	for field in requiredFields.keys():
	   if field not in existingFields:
		schemaErrors.append(dBTable+', field '+field+' is missing')
	for field in existingFields.keys():
	   if not (field in standardFields) and not (field in requiredFields):
		schemaExtensions.append(dBTable+', field '+field+' is not required')
	   # check field definition
	   if field in requiredFields.keys():
		# field type
		if existingFields[field].Type <> requiredFields[field][1]:
			schemaErrors.append(dBTable+', field '+field+', type should be '+requiredFields[field][1])
		# field nullable?
		if existingFields[field].IsNullable:
			nullStatus = 'NullsOK'
		else:
			nullStatus = 'NoNulls'
		if nullStatus <> requiredFields[field][2]:
			schemaErrors.append(dBTable+', field '+field+' should be '+requiredFields[field][2])

def loadTableValues(tableName,fieldName,valueList):
	try:
		rows = gp.searchcursor(tableName,'','',fieldName)
	except:
		loadValuesFlag = False
	else: 
		row = rows.next()
		while row:
			if row.getValue(fieldName) <> None:
				valueList.append(row.getValue(fieldName))
			row = rows.next() 
		loadValuesFlag = True
	return loadValuesFlag

def inventoryValues(gp,thisDatabase,table):
	# build list of fields in this table that must have values defined in Glossary
	# build list of fields that point at DataSources (name cn 'Source'
	# note if MapUnit field is present
	# calculate value of _ID field
	tableFields = gp.ListFields(table)
	fields = []
	mapUnitExists = False
	if debug: addMsgAndPrint(gp,'Listing fields')
	for field in tableFields:
		fields.append(field.Name)
		if field.Name == 'MapUnit':
			mapUnitExists = True
	glossFields = []
	sourceFields = []
	idField = table+'_ID'
	hasIdField = False
	if debug: addMsgAndPrint(gp,'Getting glossary and source fields')
	for field in fields:
		if field in gFieldDefList:
			glossFields.append(field)
		if field.find('Source') >= 0 and table <> 'DataSources':
			sourceFields.append(field)
		if field == idField:
			hasIdField = True
	# get rows from table of interest
	if debug: addMsgAndPrint(gp,'Getting rows')
	rows = gp.searchcursor(thisDatabase+'/'+table)
	# step through rows and inventory values in specified fields
	row = rows.next()
	while row:
		if hasIdField:
			all_IDs.append([row.GetValue(idField),table])
		if table <> 'Glossary':
			for field in glossFields:
			  try:
				if str(row.GetValue(field))[0] not in ['+','-','0','1','2','3','4','5','6','7','8','9'] or field == 'Type':
						allGlossaryRefs.append([row.GetValue(field),field,table])
			  except:
				errString = 'Table '+table+', OBJECTID = '+str(row.GetValue('OBJECTID'))
				errString = errString+', field '+field+' caused an error'
				addMsgAndPrint(gp,errString)
				addMsgAndPrint(gp,'Field value is <'+str(row.GetValue(field))+'>')
				pass
		if mapUnitExists and table <> 'DescriptionOfMapUnits' and table <> 'StandardLithology':
			allMapUnitRefs.append([row.GetValue('MapUnit'),table])
		for field in sourceFields:
			allDataSourcesRefs.append([row.GetValue(field),field,table])
		# check for pseudonulls and trailing spaces
		try:
		  tableRow = '    Table '+table+',row OBJECTID='+str(row.GetValue('OBJECTID'))+', field'
		  badNull = False
		  for field in tableFields:
			if field.Type == 'String' and not field.IsNullable:
			   #print table+'  '+field.Name
			   if row.GetValue(field.Name)[-1:] == ' ':
				tableRow = tableRow+' '+field.Name
				badNull = True
		  if badNull:
			allBadNulls.append(tableRow)
		except:
		  pass

		row = rows.next()
			
def inventoryWorkspace(gp,tables,fdsfc):
	addMsgAndPrint(gp,'  Inventorying geodatabase...')
	featureDataSets = gp.ListDatasets()
	gdbDescription.append('Tables: ')
	for table in tables:
		listDataSet(table)
	for featureDataSet in featureDataSets:
		gdbDescription.append('Feature data set: '+featureDataSet)
		gp.workspace = thisDatabase
		gp.workspace = featureDataSet
		featureClasses = gp.ListFeatureClasses()
		featureClassList = []
		for featureClass in featureClasses:
			listDataSet(featureClass)
			featureClassList.append(featureClass)
		fdsfc.append([featureDataSet,featureClassList])

def checkFieldsAndFieldDefinitions(gp):
	addMsgAndPrint(gp, '  Checking fields and field definitions, inventorying special fields...')
	for table in tables:
		if debug: addMsgAndPrint(gp,'Table = '+table)
		gp.workspace = thisDatabase
		if not tableDict.has_key(table): 
			schemaExtensions.append('Table '+table+' is not required')
		else: 
			checkTableFields(table,table)
		inventoryValues(gp,thisDatabase,table)
	for fds in fdsfc:
		gp.workspace = thisDatabase
		gp.workspace = fds[0]
		for featureClass in fds[1]:
			if debug: addMsgAndPrint(gp,'Feature class = '+featureClass)
			if not tableDict.has_key(featureClass): 
				schemaExtensions.append('Feature class '+featureClass+' is not required')
			else: 
				checkTableFields(featureClass,featureClass)
			inventoryValues(gp,thisDatabase,featureClass)

def checkRequiredElements(gp):
	addMsgAndPrint(gp, '  Checking for required elements...')
	requiredFeatureDataSets = ['GeologicMap']
	requiredTables = ['DescriptionOfMapUnits','Glossary','DataSources','StandardLithology']
	for tb1 in requiredTables:
		isPresent = False
		for tb2 in tables:
			if tb1 == tb2:
				isPresent = True
		if not isPresent:
			schemaErrors.append('Table '+tb1+' is missing')
	for fds1 in requiredFeatureDataSets:
		isPresent = False
		for fds2 in fdsfc:
			if fds2[0] == fds1: 
				isPresent = True
		if not isPresent:
			schemaErrors.append('Feature data set '+fds1+' is missing')
	for xx in fdsfc:
		fds = xx[0]
		fcs = xx[1]
		if fds[0:12] == 'CrossSection':
			checkMapFeatureClasses(gp,fds,'CS'+fds[12:],fcs)
		if fds == 'GeologicMap':
			checkMapFeatureClasses(gp,fds,'',fcs)
		if fds == 'CorrelationOfMapUnits':
			if not ('CMULines' in fcs):
				schemaErrors.append('Feature data set '+fds+', feature class CMULines is missing')
			if not ('CMUPolys' in fcs):
				schemaErrors.append('Feature data set '+fds+', feature class CMUMapUnitPolys is missing')
			if not ('CMUText' in fcs):
				schemaErrors.append('Feature data set '+fds+', feature class CMUText is missing')

def checkContent(gp):
	gp.workspace = thisDatabase
	# Check for uniqueness of _ID values
	all_IDs.sort()
	duplicateIDs.append('  Duplicate _ID values')
	for n in range(1,len(all_IDs)):
		if all_IDs[n-1][0] == all_IDs[n][0]:
			duplicateIDs.append('    '+all_IDs[n][0]+', tables '+all_IDs[n-1][1]+' '+all_IDs[n][1])
	# Check for OwnerIDs and ValueLinkIDs in ExtendedAttributes that don't match an existing _ID
	if not loadTableValues('ExtendedAttributes','OwnerID',extendedAttribIDs):
		unreferencedIds.append('    Error: did not find field OwnerID in table ExtendedAttributes')
	if not loadTableValues('ExtendedAttributes','ValueLinkID',extendedAttribIDs):
		unreferencedIds.append('    Error: did not find field ValueLinkID in table ExtendedAttributes')
	#compare
	extendedAttribIDs.sort()
	all_IDs0 = []
	for id in all_IDs:
		all_IDs0.append(str(id[0]))
	for id in extendedAttribIDs:
		if id <> None and not id in all_IDs0:
			unreferencedIds.append('    '+id)
	# Check allDataSourcesRefs against DataSources
	if loadTableValues('DataSources','DataSources_ID',dataSourcesIDs):
		# compare 
		allDataSourcesRefs.sort()
		lastID = ''
		for id in allDataSourcesRefs:
			if id[0] <> None and id[0] <> lastID:
				lastID = id[0]
				if id[0] not in dataSourcesIDs:
					missingSourceIDs.append('    '+id[0]+', cited in field '+id[1]+' table '+id[2])
		# compare other way
		allSourceRefIDs = []
		for ref in allDataSourcesRefs:
			allSourceRefIDs.append(ref[0])
		for id in dataSourcesIDs:
			if not id in allSourceRefIDs:
				unusedDataSources.append('    '+id)
	else:
		missingSourceIDs.append('    Error: did not find field DataSources_ID in table DataSources')
		unusedDataSources.append('    Error: did not find field DataSources_ID in table DataSources')
	# Check MapUnits against DescriptionOfMapUnits and StandardLithology
	if not loadTableValues('DescriptionOfMapUnits','MapUnit',dmuMapUnits):
		missingDmuMapUnits.append('    Error: did not find field MapUnit in table DescriptionOfMapUnits')
		unreferencedDmuMapUnits.append('    Error: did not find field MapUnit in table DescriptionOfMapUnits')
	if not loadTableValues('StandardLithology','MapUnit',standardLithMapUnits):
		missingStandardLithMapUnits.append('    Error: did not find field MapUnit in table StandardLithology') 
		unreferencedStandardLithMapUnits.append('    Error: did not find field MapUnit in table StandardLithology')
	# compare 
	allMapUnitRefs.sort()
	lastMU = ''
	allMapUnitRefs2 = []
	for mu in allMapUnitRefs:
		if mu[0] <> lastMU:
			lastMU = mu[0]
			allMapUnitRefs2.append(lastMU)
			if mu[0] not in dmuMapUnits:
				missingDmuMapUnits.append('    '+mu[0]+', cited in '+mu[1])
			if mu[0] not in standardLithMapUnits:
				missingStandardLithMapUnits.append('    '+mu[0]+', cited in '+mu[1])
	# look for excess map units in StandardLithology
	lastMu = ''
	standardLithMapUnits.sort()
	for mu in standardLithMapUnits:
		if mu <> lastMu and not mu in allMapUnitRefs2:
			lastMu = mu
			unreferencedStandardLithMapUnits.append('    '+mu)
	# look for unreferenced map units in DMU
	for mu in dmuMapUnits:
		if mu not in allMapUnitRefs2 and mu <> '':
			unreferencedDmuMapUnits.append('    '+mu)
	# Check allGlossaryRefs against Glossary
	if loadTableValues('Glossary','Term',glossaryTerms):
		# compare 
		allGlossaryRefs.sort()
		lastTerm = ''
		for term in allGlossaryRefs:
			if str(term[0]) <> 'None' and term[0] <> lastTerm:
				lastTerm = term[0]
				if term[0] not in glossaryTerms:
					if len(term[0]) < 40:
						thisTerm = term[0]
					else:
						thisTerm = term[0][0:37]+'...'
					missingGlossaryTerms.append('    '+thisTerm+', cited in field '+term[1]+', table '+term[2])
		# compare other direction
		glossaryRefs = []
		for ref in allGlossaryRefs:
			glossaryRefs.append(str(ref[0]))
		for term in glossaryTerms:
			if term not in glossaryRefs:
				unusedGlossaryTerms.append('    '+term)
	else:
		missingGlossaryTerms.append('    Error: did not find field Term in table Glossary')
		unusedGlossaryTerms.append('    Error: did not find field Term in table Glossary')
	# Check GeologicEvents against ValueLinkID in ExtendedAttributes
	geologicEvents = []
	if not loadTableValues('GeologicEvents','GeologicEvents_ID',geologicEvents):
		unusedGeologicEvents.append('    Error: did not find field GeologicEvents_ID in table GeologicEvents')
	valueLinks = []
	if not loadTableValues('ExtendedAttributes','ValueLinkID',valueLinks):
		unusedGeologicEvents.append('    Error: did not find field ValueLink in table ExtendedAttributes')
	#compare
	geologicEventRefs = []
	for ve in valueLinks:
		if ve in geologicEvents:
			geologicEventRefs.append(ve)

	# Check formatting of HierarchyKey in DescriptionOfMapUnits
	hKeys = []
	if loadTableValues('DescriptionOfMapUnits','HierarchyKey',hKeys):
		partLength = len(hKeys[0].split('-')[0])
		for hKey in hKeys:
			hKeyParts = hKey.split('-')
			keyErr = False
			for hKeyPart in hKeyParts:
				if len(hKeyPart) <> partLength:
					keyErr = True
			if keyErr:
				hKeyErrors.append('    '+hKey)	
	else:
		hKeyErrors.append('    Error: did not find field HierarchyKey in table DescriptionOfMapUnits')

def writeOutput(gp,outFile,thisDatabase):
	addMsgAndPrint(gp, '  Writing output...')
	outfl = open(outFile,'w')
	outfl.write('Geodatabase '+thisDatabase+'\n')
	outfl.write('  Testing for compliance with NCGMP09 database schema\n')
	outfl.write('  This file written by '+versionString+'\n')
	outfl.write('  '+time.asctime(time.localtime(time.time()))+'\n')
	#  Schema errors
	outfl.write('\n\nSCHEMA ERRORS\n\n')
	if len(schemaErrors) == 0:
		outfl.write('  None\n')
	else: 
		for aline in schemaErrors:
			outfl.write('  '+aline+'\n')
	# Extensions to schema
	outfl.write('\n\nEXTENSIONS TO SCHEMA, may indicate errors\n\n')
	if len(schemaExtensions) == 0:
		outfl.write('  None\n')
	else: 
		for aline in schemaExtensions:
			outfl.write('  '+aline+'\n')
	# Content errors
	outfl.write('\n\nCONTENT ERRORS\n\n')
	writeContentErrors(outfl,duplicateIDs,'No duplicate _IDs')
	writeContentErrors(outfl,missingSourceIDs,'No missing entries in DataSources')
	writeContentErrors(outfl,unusedDataSources,'No unreferenced entries in DataSources')
	writeContentErrors(outfl,missingDmuMapUnits,'No missing MapUnits in DescriptionOfMapUnits')
	writeContentErrors(outfl,missingStandardLithMapUnits,'No missing MapUnits in StandardLithology')
	writeContentErrors(outfl,unreferencedDmuMapUnits,'No unreferenced MapUnits in Description of MapUnits')
	writeContentErrors(outfl,unreferencedStandardLithMapUnits,'No unreferenced MapUnits in StandardLithology')
	writeContentErrors(outfl,missingGlossaryTerms,'No missing terms in Glossary')
	writeContentErrors(outfl,unusedGlossaryTerms,'No unreferenced terms in Glossary')
	writeContentErrors(outfl,unreferencedIds,'No rows in ExtendedAttributes that reference nonexistent OwnerIDs or ValueLinkIDs')
	writeContentErrors(outfl,unusedGeologicEvents,'No rows in GeologicEvents not referenced in ExtendedAttributes')
	writeContentErrors(outfl,hKeyErrors,'No format errors in HierarchyKeys')
	writeContentErrors(outfl,allBadNulls,'No pseudonulls or trailing spaces')
	# Database description
	outfl.write('\nGEODATABASE DESCRIPTION\n\n')
	for aline in gdbDescription:
		outfl.write(aline+'\n')
	outfl.close()

def validInputs(gp,thisDatabase,outFile):
	# does input database exist? Is it plausibly a geodatabase?
	if os.path.exists(thisDatabase) and thisDatabase[-4:].lower() in ('.gdb','.mdb'):
		# is output workspace writable?
		try:
			outfl = open(outFile,'w')
			outfl.write('A test')
			outfl.close()
			return True
		except:
			addMsgAndPrint(gp,'  Cannot open and write file '+outFile)
			return False
	else:
		addMsgAndPrint(gp,'  Object '+thisDatabase+' does not exist or is not a geodatabase')
		return False


thisDatabase = sys.argv[1]
outputWorkspace = sys.argv[2]
print '  '+versionString
try:
	print '  Creating geoprocessing object...'
	gp = arcgisscripting.create(9.3)
except:
	print '  Unable to create geoprocessing object. Arc license not present??'
else:
  if not outputWorkspace[-1:] in ('/','\\'):
	outputWorkspace = outputWorkspace+'/'
  outFile = outputWorkspace + os.path.basename(thisDatabase)+'-conformance.txt'
  thisDatabase = os.path.abspath(thisDatabase)
  if validInputs(gp,thisDatabase,outFile):
	try:
		gp.workspace = thisDatabase
	except:
		addMsgAndPrint(gp,'  Unable to load workspace '+thisDatabase+'. Not an ESRI geodatabase?')
	else:
		addMsgAndPrint(gp,'  '+versionString)
		addMsgAndPrint(gp,'  geodatabase '+thisDatabase+' loaded')
		addMsgAndPrint(gp,'  output will be written to file '+outFile)
		tables = list(gp.ListTables())
		inventoryWorkspace(gp,tables,fdsfc)
		checkRequiredElements(gp)
		checkFieldsAndFieldDefinitions(gp)
		checkContent(gp)
		writeOutput(gp,outFile,thisDatabase)
		addMsgAndPrint(gp,'  DONE')


# Things to be done:
#	Inventory data set   DONE
#	Check for required tables, featuredatasets, featureclasses    DONE 
#	For optional featuredatasets, check for required featureclasses   DONE
#	For required and optional tables, 
#		check for required fields   DONE
#		check field definitions    DONE
# 	Accumulate list of all must-be-globally-unique (_ID) values  DONE
#	Accumulate list of all ToBeDefined values (references to Glossary)   DONE
#	   see gFieldDefList, above
#	   what about StandardLithology:Lithology?
#	Accumulate list of all MapUnit references  DONE
# 	Accumulate list of all DataSources references  DONE
#	Check that must-be-globally-unique values are unique  DONE
#	Check that all values of fields name cn 'Source' correspond to values of DataSources_ID  DONE
#	Check that MapUnits referenced by StandardLithology and any fClass whose name cn 'Poly' 
#		are present in DMU  DONE
#	Check for definitions of ToBeDefined values  DONE
#        EXCEPT PropertyValue values that are numeric (in part)  STILL TO BE DONE!
#	In DMU, check HierarchyKey format  DONE
#	Check for illegal null values: ' ' (null-equivalent) created when loading data into tables  DONE
#	Check that all MapUnits have StandardLithology entries  DONE
#	Check for unreferenced entries in DataSources (DONE), Glossary (DONE), StandardLithology (DONE),
#		DMU (DONE), Geologic Events (DONE),
# 	DOES NOT CHECK THAT GeologicEvents REFERENCED BY ValueLinkID IN ExtendedAttributes ARE PRESENT
#	Error trapping:
#		So it doesn't fail with bad input  DONE, somewhat
#		So it doesn't fail with bad field names. e.g. _ID field  DONE, I think
